//
// f4500_root.cc
//    Ver. 4.00 (2021/08/17)
//    Copyright by Y.Tajima
//


#include <iostream>
#include <string>
#include <vector>
#include <sstream>
#include <cstdlib>

#include <string.h>
#include <math.h>
#include <unistd.h>

// root
#include "TFile.h"
#include "TH1D.h"
#include "TH2D.h"

#include "csv.hh"

const char * BASEDIR="./";     // default home directory
const char * CSVDIR= "csv/";   // csv directory  (UTF8)
const char * ROOTDIR="root/";  // root directory (Output)

#define MAXWL 10

using namespace std;

// copy from string to double/integer variable
double copyDouble(string str) {
  stringstream ss;
  double ret;
  ss << str;  ss >> ret;
  return ret;
}
int copyInt(string str) {
  stringstream ss;
  int ret;
  ss << str; ss >> ret;
  return ret;
}
// reflection light cut
bool refCut (double exWave, double emWave, double width) {
  return (abs(exWave    - emWave) > width &&     // 反射光
          abs(exWave*2. - emWave) > width &&     // 2倍波長光
          abs(exWave*3. - emWave) > width );     // 3倍波長光 
}
// point cut
bool pointCut (double exWave, double emWave,  Int_t *pEx, Int_t *pEm, Int_t numPoint) {
  bool ret = true;
  for (Int_t i = 0; i < numPoint; i++) {
    if ((Int_t)exWave == pEx[i] && (Int_t)emWave == pEm[i]) { ret = false; }
  }
  return ret;
}

//
// main program
//
int main(int argc, char** argv) {
  // option extraction
  int opt;
  int numEx = 0, numEm = 0, numPoint = 0;
  int wlEx[MAXWL], wlEm[MAXWL], pEx[MAXWL], pEm[MAXWL]; 
  int cutFlag = 0, errFlag = 0, peakFlag = 0, aveFlag = 0, twoFlag = 0;
  int lowEx = 0, lowEm = 0, aveWl = 0;
  double wFactor = 1., twoCut = 2.0;
  string cutWL, pointWL;
  int sep;
  char *subDir = NULL;

  while ((opt = getopt(argc, argv, "x:m:d:l:o:c:2:a:p")) != -1)  {
    switch (opt) {
    case 'x': // -x exWaveLength : make 1D histogram with fixed excitation wave length
      wlEx[numEx++] = atoi(optarg);
      break;
    case 'm': // -m emWaveLength : make 1D histogram with fixed emission wave length
      wlEm[numEm++] = atoi(optarg);
      break;
    case 'd': // -d directory : sub directory
      subDir = optarg;
      break;
    case 'l': // -l lowLimitExciteWaveLength[:lowLomitEmissionWaveLength] : cut waveklength of low limit
      cutWL = optarg;
      lowEx = atoi(optarg);
      sep = cutWL.find(":");
      if (sep != string::npos) {
        cutWL.erase(0,sep+1);
        lowEm = atoi(cutWL.c_str());
      }
      break;
    case 'o': // -o exciteWaveLength:emissionWaveLength : cut point data in 2D histogram
      pointWL = optarg;
      pEx[numPoint] = atoi(optarg);
      sep = pointWL.find(":");
      pointWL.erase(0, sep+1);
      pEm[numPoint++] = atoi(pointWL.c_str());
      break;
    case 'c': // -c widthFactor : cut reflectioln light in 2D histogram
      cutFlag = 1;
      wFactor = atof(optarg);
      break;
    case '2': // -2 cutFactor : cut (ex * cutFactor < em) area  
      twoFlag = 1;
      twoCut = atof(optarg);
      break;
    case 'a': // -a aveWaveLength : cut average edge (Ex + Em < aveWl*2)
      aveFlag = 1;
      aveWl = atoi(optarg);
      break;
    case 'p': // -p : auto create 1D histograms at peak channel
      peakFlag = 1; 
      break;
    default: 
      errFlag = 1;
    }
  }
  if (errFlag == 1 || optind >= argc) {
    cerr << "Usage:" << argv[0] << " [-d subDirectoryName] [-x ExcitedWaveLength] [-m EmittedWaveLength]"
    << " [-c widthFactor] [-l lowCutExcitationWL[:lowCutEmissionWL]]"
    << " [-o cutExcitationWL:cutEmissionWL] [-a aveWaveLength] [-p] <Filename> " << endl;
    return EXIT_FAILURE;
  }

  // get environment variables (F4500_BASE, F4500_CSV, F4500_ROOT)
  const char * c_base = getenv("F4500_BASE");
  const char * c_csv  = getenv("F4500_CSV");
  const char * c_root = getenv("F4500_ROOT");

  if (c_base == NULL) { c_base=BASEDIR; }
  if (c_csv  == NULL) { c_csv =CSVDIR;  }
  if (c_root == NULL) { c_root=ROOTDIR; }
  string basedir = c_base;
  string csvdir  = c_csv;
  string rootdir = c_root;
 
  string csvfile,rootfile;
  if (subDir == NULL) {
    csvfile  = basedir + csvdir + argv[optind] + ".csv";
    rootfile = basedir + rootdir + argv[optind] + ".root";
  } else {
    csvfile  = basedir + csvdir  + subDir + "/" + argv[optind] + ".csv";
    rootfile = basedir + rootdir + subDir + "/" + argv[optind] + ".root";
  }

  // read csv file 
  cerr << csvfile << endl;  
  vector<string> csvData;
  Csv objCsv;

  if (!objCsv.openFile(csvfile.c_str())) {
    cerr << "file open error: " << csvfile << endl;
    return -1;
  }
 
  // make 2dim. data
  double excit[4], emit[4];
  int excitBin, emitBin, pmt;
  bool dataFlag = false;
  string title, titleCut; 
  vector<vector<double>> data;
  while (objCsv.readLine(csvData)) { // read & split data in 1 line
    // remove double-quotation
    for (int col=0; col< csvData.size(); col++) {
      if (csvData[col][0] == '"') {
        csvData[col].erase(csvData[col].begin());
        csvData[col].pop_back();
      }
    }
    if (csvData[0] == "ファイル名:") {
      title = csvData[1]; title.erase(title.length()-4, 4);
      titleCut = title + "Cut";
    }
    if (csvData[0] == "ホトマル電圧:")   { pmt = copyInt(csvData[1]); }
    if (csvData[0] == "励起開始波長:")   { excit[0] = copyDouble(csvData[1]); }
    if (csvData[0] == "励起終了波長:")   { excit[1] = copyDouble(csvData[1]); }
    if (csvData[0] == "励起側スリット:") { excit[3] = copyDouble(csvData[1]); }
    if (csvData[0] == "励起側サンプリング間隔:") {
      excit[2] = copyDouble(csvData[1]);
      excitBin = (excit[1] - excit[0]) / excit[2] + 1;
    }
    if (csvData[0] == "蛍光開始波長:")   { emit[0] = copyDouble(csvData[1]); }
    if (csvData[0] == "蛍光終了波長:")   { emit[1] = copyDouble(csvData[1]); }
    if (csvData[0] == "蛍光側スリット:") { emit[3] = copyDouble(csvData[1]); }
    if (csvData[0] == "蛍光側サンプリング間隔:") {
      emit[2] = copyDouble(csvData[1]);
      emitBin = (emit[1] - emit[0]) / emit[2] + 1;

      data.resize(emitBin);
      for (int em=0; em < emitBin; em++) { data[em].resize(excitBin); }
    }

    if (dataFlag && csvData[0].length() > 2) { // 3桁以上の数字(発光波長)
      int em = (copyInt(csvData[0]) - emit[0]) / emit[2];
      for (int ex = 0; ex < excitBin; ex++) {
        data[em][ex] = copyDouble(csvData[ex+1]);
      }
      //      cout << csvData[0] << " " << csvData[1] << " " << em << endl; 
    }
    if (csvData[0] == "データリスト") dataFlag = true;
    csvData.clear();
  }

  // print parameters
  cerr << "Title:   " <<  title << endl;
  cerr << "Excit:   " << excit[0] << "-" << excit[1] << " Step: " << excit[2] << " Slit: " << excit[3] << " Bin: " << excitBin <<  endl;
  cerr << "Emit:    " << emit[0]  << "-" << emit[1]  << " Step: " << emit[2]  << " Slit: " << emit[3]  << " Bin: " << emitBin  << endl;
  
  
  // make root file
  TFile *rootFile  = new TFile(rootfile.c_str(), "recreate");
  TH2D *hist2d, *hist2dcut;

  //   make 2d histogram  
  hist2d = new TH2D(title.c_str(), title.c_str(),
                    emitBin,  emit[0]-emit[2]/2.,   emit[1]+emit[2]/2.,
                    excitBin, excit[0]-excit[2]/2., excit[1]+excit[2]/2.);
  hist2dcut = new TH2D(titleCut.c_str(),titleCut.c_str(), 
                    emitBin,  emit[0]-emit[2]/2.,   emit[1]+emit[2]/2.,
                    excitBin, excit[0]-excit[2]/2., excit[1]+excit[2]/2.);

  double width  = sqrt(excit[2]*excit[2] + emit[2]*emit[2]) * wFactor;

  for (int em = 0; em < emitBin; em++) {
    for (int ex = 0; ex <excitBin; ex++) {
      double exWave = excit[0] + ex * excit[2];
      double emWave = emit[0]  + em * emit[2];
      hist2d->SetBinContent(em+1, ex+1, data[em][ex]);
      if ( exWave < emWave && exWave >= lowEx && emWave >= lowEm &&   // 短波長ノイズカット
           (aveFlag == 0 || exWave + emWave > aveWl*2) &&             // 短波長斜めカット
           (numPoint == 0 || pointCut(exWave, emWave, pEx, pEm, numPoint)) && // ポイントカット
           (twoFlag == 0 || exWave * twoCut - emWave > 0) &&               // 2倍光以上カット 
           (cutFlag == 0 || refCut(exWave, emWave, width))) {         // 反射光カット
        hist2dcut->SetBinContent(em+1, ex+1, data[em][ex]);
        //        cout << em << " " << ex << " " << data[em][ex] << endl; 
      }
    }
  }
  
  if (peakFlag == 1) {
    Int_t exBin, emBin, iz;
    hist2dcut->GetBinXYZ(hist2dcut->GetMaximumBin(), emBin, exBin, iz);
    Int_t exWave = excit[0] + (exBin -1) * excit[2];
    Int_t emWave = emit[0]  + (emBin -1) * emit[2];    
    wlEx[numEx++] = exWave;
    wlEm[numEm++] = emWave;
  }

  // Projection 1d histograms
  TH1D *histX, *histY;
  string titleX = title+"_Em";
  histX = hist2dcut->ProjectionX(titleX.c_str());
  string titleY = title+"_Ex";
  histY = hist2dcut->ProjectionY(titleY.c_str());

  // make 1d histograms
  TH1D *histEx[MAXWL], *histEm[MAXWL];
  //   Fixed Excitation
  for (int i = 0; i < numEx; i++) {
    cerr << "Excit1D: " << wlEx[i] << "nm"  << endl;
    int ex = (wlEx[i] - excit[0]) / excit[2];
    string title1d = title + "_Ex" + to_string(wlEx[i]);
    histEx[i] = new TH1D(title1d.c_str(), title1d.c_str(), 
                         emitBin, emit[0]-emit[2]/2., emit[1]+emit[2]/2.);
    for (int em = 0; em < emitBin; em++) {
      double emWave = emit[0] + em * emit[2];
      if ( wlEx[i] < emWave && emWave >= lowEm 
            && (aveFlag == 0 || wlEx[i] + emWave > aveWl*2)
            && (cutFlag == 0 || refCut(wlEx[i], emWave, width))) {
        histEx[i]->SetBinContent(em, data[em][ex]);
      }
    }
  }
  //   Fixed Emission
  for (int i = 0; i < numEm; i++) {
    cerr << "Emit1D:  " << wlEm[i] << "nm"  << endl;
    int em = (wlEm[i] - emit[0]) / emit[2];
    string title1d = title + "_Em" + to_string(wlEm[i]);
    histEm[i] = new TH1D(title1d.c_str(), title1d.c_str(),
                         excitBin, excit[0]-excit[2]/2., excit[1]+excit[2]/2.);
    for (int ex = 0; ex < excitBin; ex++) {
      double exWave = excit[0] + ex * excit[2];
      if ( exWave < wlEm[i] && exWave >= lowEx 
            && (aveFlag == 0 || exWave + wlEm[i] > aveWl)
            && (cutFlag == 0 || refCut(exWave, wlEm[i], width))) { 
        histEm[i]->SetBinContent(ex, data[em][ex]);
      }
    }
  }

  rootFile->Write();
  rootFile->Close();
 
  // data print
  /*
  for (int em = 0; em < emitBin; em++) {
    for (int ex = 0; ex < excitBin; ex++) {
      cout << data[em][ex] << " " ;
    }
    cout << endl;
  }
  */
  
  return 0;
}
